##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
import time, string
_marker = []
from Persistence import Persistent
import Acquisition
from zLOG import LOG

class AutoExpireMapping:
    """
    Pure mixin to a session data container or for subclassing.
    
    A mapping which self-manages the expiration of the objects
    it contains based on user-provided timeout value and resolution
    by using a ring of timed buckets.
    """

    def __init__(self, timeout_mins=20, ctype=None, onend='onEnd',
                 err_margin=.20):
        # we give ourselves a 20% error margin by default to reduce
        # the number of buckets necessary
        self._setTimeout(timeout_mins)
        if ctype is None:
            LOG('Session Tracking', 100,
                ('AutoExpireMapping container type set to dictionary... this'
                 ' is likely an error'))
            def ctype():
                return {}
        self._ctype = ctype
        self._setOnEnd(onend)
        self._err_margin = err_margin
        self._reset()
        
    def getTimeoutMinutes(self):
        """ """
        return self._timeout_secs / 60

    def _setTimeout(self, timeout_mins):
        # incoming timeout value is in minutes
        if type(timeout_mins) is not type(1):
            raise "Timeout value must be integer: %s" % timeout_mins
        # convert to seconds and store, timeout must be stored as integer
        self._timeout_secs = timeout_mins * 60
        
    def _setOnEnd(self, onend):
        if type(onend) is not type(''):
            raise "Onend function name must be string, %s" % onend
        if onend[0:1] not in string.letters:
            raise "Onend function name must start with a-z, %s" % onend
        self._onend = onend

    def _reset(self):
        self._resolution_secs = int(self._timeout_secs*self._err_margin) or 1
        now = int(time.time())
        ctype = self._ctype
        numbuckets = self._timeout_secs/self._resolution_secs or 1
        l = []
        i = 0
        for x in range(numbuckets):
            dump_after = now + i
            c = ctype()
            l.insert(0, [c, dump_after])
            i = i + self._resolution_secs
        index = ctype()
        self._ring = Ring(l, index)

    def _getCurrentBucket(self, get_dump=0, time=time.time):
        # no timeout always returns last bucket
        if not self._timeout_secs:
            b, dump_after = self._ring._data[0]
            return b
        index = self._ring._index
        now = int(time())
        i = self._timeout_secs
        # expire all buckets in the ring which have a dump_after time that
        # is before now, turning the ring as many turns as necessary to
        # get to a non-expirable bucket.
        while 1:
            l = b, dump_after = self._ring._data[-1]
            if now > dump_after:
                self._ring.turn()
                # mutate elements in-place in the ring
                new_dump_after = now + i
                l[1] = new_dump_after
                self._clean(b, index)
                i = i + self._resolution_secs
            else:
                break
        if get_dump:
            return self._ring._data[0], dump_after, now
        else:
            b, dump_after = self._ring._data[0]
            return b

    def _clean(self, b, index):
        for ob in b.values():
            d = last = None
            f = getattr(ob, self._onend, None)
            getDataMgr = getattr(ob, 'getDataMgr', None)
            if getDataMgr is not None:
                if callable(getDataMgr):
                    d = getDataMgr()
                if d != last:
                    mgr = self.aq_parent.unrestrictedTraverse(d)
                    last = d
            if callable(f): f(mgr)
        for k, v in list(index.items()):
            if v is b: del index[k]
        b.clear()
        
    def _show(self):
        """ debug method """
        b,dump,now = self._getCurrentBucket(1)
        ringdumps = map(lambda x: `x[1]`[-4:], self._ring)
        t = (
            "now: "+`now`[-4:],
            "dump_after: "+`dump`[-4:],
            "ring_dumps: "+`ringdumps`,
            "ring: " + `self._ring`
            )
        for x in t:
            print x

    def __setitem__(self, k, v):
        current = self._getCurrentBucket()
        index = self._ring._index
        b = index.get(k)
        if b is None:
            # this is a new key
            index[k] = current
        elif b is not current:
            # this is an old key that isn't in the current bucket.
            del b[k] # delete it from the old bucket
            index[k] = current
        # change the value
        current[k] = v
        
    def __getitem__(self, k):
        current = self._getCurrentBucket()
        index = self._ring._index
        # the next line will raise the proper error if the item has expired
        b = index[k]
        v = b[k] # grab the value before we potentially time it out.
        if b is not current:
            # it's not optimal to do writes in getitem, but there's no choice.
            # we accessed the object, so it should become current.
            index[k] = current # change the index to the current bucket.
            current[k] = v # add the value to the current bucket.
            del b[k] # delete the item from the old bucket.
        return v

    def get(self, k, default=_marker):
        try: v = self[k]
        except KeyError: v = _marker
        if v is _marker:
            if default is _marker:
                return None
            else:
                return default
        return v
        
    def __delitem__(self, k):
        self._getCurrentBucket()
        index = self._ring._index
        b = index[k]
        del index[k]
        del b[k]

    def __len__(self):
        self._getCurrentBucket()
        return len(self._ring._index)

    def has_key(self, k):
        self._getCurrentBucket()
        index = self._ring._index
        return index.get(k, _marker) is not _marker

    def values(self):
        return map(lambda k, self=self: self[k], self.keys())

    def items(self):
        return map(lambda k, self=self: (k, self[k]), self.keys())

    def keys(self):
        self._getCurrentBucket()
        index = self._ring._index
        return map(lambda x: x, index.keys())

    def update(self):
        raise "Unsupported"

    def clear(self):
        raise "Unsupported"

    def copy(self):
        raise "Unsupported"

    getLen = __len__
    
class Ring(Persistent):
    """ Instances of this class will be frequently written to the ZODB,
    so it's optimized as best possible for write-friendliness """
    def __init__(self, l, index):
        if not len(l):
            raise "ring must have at least one element"
        self._data = l
        self._index = index

    def __repr__(self):
        return repr(self._data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, i):
        return self._data[i]

    def turn(self):
        last = self._data.pop(-1)
        self._data.insert(0, last)
        self._p_changed = 1

    def _p_independent(self):
        return 1
    
class PersistentAutoExpireMapping(AutoExpireMapping, Persistent,
                                  Acquisition.Explicit):
    """ Used by SessionFanout """
    pass
    
