##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

'''Patches the ObjectManager class so it checks for attributes in
transparent folders when they exist.  Any object with the attribute
_is_transparent will be searched for extra attributes.
'''

from OFS.ObjectManager import ObjectManager
from Globals import Persistent, Acquisition
from ExtensionClass import Base


class Rewrapper (Base):
    def __init__(self, path):
        self._path = path

    def __of__(self, parent):
        ob = parent
        for p in self._path:
            ob = getattr(ob, p)
        return ob


try:
    # getOtherAttr is assigned to whatever ObjectManager.__getattr__
    # used to do.
    try:
        oldGetAttr = ObjectManager.__getattr__
    except:
        try:
            oldGetAttr = ObjectManager.inheritedAttribute('__getattr__')
        except:
            def oldGetAttr(self, name):
                raise AttributeError, name

    def _getattr(self, name):
        # This is the new ObjectManager.__getattr__().  If an object by the
        # given ID is not found through the normal mechanisms, we
        # fall through to here and try to find the object in a
        # transparent folder.
        tpids = self._transparent_object_ids
        if tpids and not self._v_no_transparent and name[:3] != '_p_':
            # The attribute may be accessible through
            # transparent folders.
            try:
                # Avoid recursion.
                self._v_no_transparent = 1
                for tpid in tpids:
                    tp = self.__dict__.get(tpid, None)
                    if hasattr(tp, name):
                        # Return an object that will, through a call
                        # to __of__(), manifest itself as the wrapped
                        # object.
                        return Rewrapper((tpid, name))
            finally:
                self._v_no_transparent = 0
        return oldGetAttr(self, name)

    def _checkId(self, id, allow_dup=0):
        # Replacement for ObjectManager._checkId().
        if not allow_dup and self._transparent_object_ids:
            # If the object can only be found in a transparent folder,
            # it's not really here and can be 'duplicated'.
            try:
                self._v_no_transparent = 1
                base = getattr(self,  'aq_base', self)
                if not hasattr(base, id):
                    # Cause _checkId to not check for duplication.
                    return self._old_checkId(id, allow_dup=1)
            finally:
                self._v_no_transparent = 0
        return self._old_checkId(id, allow_dup)

    def _setOb(self, id, object):
        # Besides performing setattr(), adds or removes the object
        # to/from the list of transparent subfolders.
        self._old_setOb(id, object)
        isTransparent = getattr(object, '_is_transparent', 0)
        _updateTransparentId(self, id, isTransparent)

    def _delOb(self, id):
        # Besides performing delattr(), removes the object
        # from the list of transparent subfolders.
        _updateTransparentId(self, id, 0)
        self._old_delOb(id)

    def _updateTransparentId(self, name, isTp):
        # Puts an object id in the list of _transparent_ids or removes it.
        if isTp:
            # Add to list.
            if name not in self._transparent_object_ids:
                idlist = tuple(self._transparent_object_ids) + (name,)
                self._transparent_object_ids = idlist
        elif self._transparent_object_ids:
            # Remove from list.
            try:
                self._transparent_object_ids = filter(
                    lambda id, name=name: id != name,
                    self._transparent_object_ids)
            except:
                # Already removed.
                pass

    # Apply the patch.
    print 'Applying TransparentFolders patch version 0.3.'
    ObjectManager._transparent_object_ids = ()
    ObjectManager._v_no_transparent = 0
    ObjectManager._old_checkId = ObjectManager._checkId
    ObjectManager._checkId = _checkId
    ObjectManager._old_setOb = ObjectManager._setOb
    ObjectManager._setOb = _setOb
    ObjectManager._old_delOb = ObjectManager._delOb
    ObjectManager._delOb = _delOb
    ObjectManager.__getattr__ = _getattr

except:
    import traceback
    traceback.print_exc()
